#! /bin/sh

# This script assumes that all hosts that run the node managers and system
# managers share a common filesystem. If this is not the case, run setup
# for each separately.

# Edit next line. Path name of the top of the PRM directory hierarchy. 
PRM_TOP=/nfs/pfs/divirs/jmsm/PRM

# Edit next line. Put in the name of the Host on which system manager runs.
SMHOSTS="zephyr.isi.edu"

# Next line contains the name of the file in which node names are specified.
# Node names are internet-style host names. All host names are specified on a
# single line, separated by spaces.
NODESFILE=nodes.list

NMOPTS=
SMOPTS=

# ping used to find out whether a host is on the network.
PING=/usr/etc/ping

BINDIR=$PRM_TOP/bin

# Directories in which nodemngr and sysmngr binaries reside. At execution time
# the $ARCH gets appended to this.
NMGRDIR=$BINDIR    
SMGRDIR=$BINDIR    

COUNT=`expr 0`

usage="Usage: setup [ -n \"nodemngr-opts\" ] [ -s \"sysmngr-opts\" ]"

#set -x

while [ $# -gt 0 ]; do

	case "$1" in
	-s)
		SMOPTS=$2
		shift ; shift
		;;
	-n)     
		NMOPTS=$2
		shift ; shift 
		;;
	-f)     
		NODESFILE=$2
		shift ; shift 
		;;
	-*)
		echo "$usage" 1>&2 
		exit 1
		;;
	*)
		break
		;;
	esac
done

if [ $# -gt 0 ]; then
	echo "$usage" 1>&2 
	exit 1
fi

rm -f /tmp/nm1
touch /tmp/nm1

# start nodemngrs on every NMHOST

read hostlist < $NODESFILE

for  host in $hostlist ; do
	{ { [ $ARCH = hp700 ] || $PING $host 5 > /dev/null 2>&1 ; } && \
	  rsh $host "exec /bin/sh -cx \" cd $NMGRDIR/\`arch\` ; exec nice \
	  nodemngr $NMOPTS \" < /dev/null >& /tmp/nodemngr_err.$$ " ; }
	if [ $? ] ; then
		echo "$host `rsh $host arch`" >> /tmp/nm1
                COUNT=`expr $COUNT + 1`
	else
		echo "Could not start nodemngr on $host"
	fi

#	read host NMOPTS  < $NODESFILE
done

mv -f /tmp/nm1 $SMGRDIR/`arch`/nmfile-${COUNT}.setup
rm -f /tmp/nm1

sleep 3  # allow time for nodemngrs to start up fully 

# start system manager(s)

for host in $SMHOSTS ; do
	{ { [ $ARCH = hp700 ] ||  $PING $host 5 > /dev/null 2>&1 ; } && \
	  rsh $host "exec /bin/sh -cx \" cd $SMGRDIR/\`arch\` ; exec sysmngr \
	  -f nmfile-${COUNT}.setup $SMOPTS \" < /dev/null >& \
	  /tmp/sysmngr_err.$$ " ; }
	if [ $? != 0 ] ; then
		echo "Could not reach $host to start sysmngr"
	fi
done
