/*
 * Copyright (c) 1992, 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the files
 * <prm-copyr.h>.
 *
 * Originally managers.h. Modified by srao.
 *
 */

#include <prm-copyr.h>


#ifndef _PRM_H_INCLUDED
#define _PRM_H_INCLUDED

/*  Every message contains an opcode specifying the contents of the message
    and/or type of service requested by the client. The opcode is of type 
    u_char, and is a small positive value. It is placed at the offset 
    PRM_OP_OFFSET from the start of the message (pkt->start). Multipacket
    message need to have the opcode specified only in the first packet.
    
*/

#define TASK_INIT   1  /* Initialize a task on node an schedule it for exec */
#define TINIT_SCS   2  /* Task initiation successful (sent by nodemngr) */ 
#define TASKDONE    3  /* A task completed execution */
#define JOBDONE     4  /* All tasks (and hence job) completed execution */

#define TTBL_UPDATE 5  /* Update table of Tid-hostaddr translations  */
#define HOSTID      6

#define RSC_ALLOC    16  /* Resource allocation request from jobmngr */
#define NSTATUS_QRY  17  /* Node status query from sysmngr */
#define NSTAT_UPDATE 18  /* Update availability status of node */
#define AUTHCODE     19  /* Authorization code */
#define CPY_CRNT     21  /* Local copy of file is up to date */

#define TASKCOUNT    32  /* Total number of tasks */
#define APP_MSG      33  /* Application-level message */
#define TASK_SYNC    34  /* Jobmngr enforced synchronization of tasks */
#define APPL_ERR     35
#define IO_PRINT     36  /* Print string to terminal (meant for tiotask only)
			  */
#define IJADD_REQ    48  /* Request socket/port addr of jobmngr or iotasks */
#define HADD_REQ     49  /* Request port/hostaddr of task */
#define PMAP_REQ     50  /* Request UDP port num of remote task */
#define PBINDREQ     51  /* Assign a UDP port number to socket */


/* Opcodes for file operations on remote files */

#define FILE_SEND   64

#if 0  /* These codes are not used anymore. Replaced by PDAP codes  */

#define FILE_OPEN   65
#define FILE_CLOSE  66 
#define FILE_READ   67
#define FILE_WRITE  68
#define FILE_STAT   69
#define FILE_SEEK   70
#define FILE_UNLINK 72

#endif

/* Offsets (from beginning of message text) of various fields in message. 
   Different messages use different field formats, so the same offset may
   be used for different things in different messages
*/

#define PRM_OP_OFFSET 0 /* A one byte field at this offset identifies the 
			   operation to perform  */

#define PRM_JIDOFFSET 1    /*  Messages sent by the job manager and the tasks 
			   always carry a jobid at the beginning */

#define OSCODE_OFFSET 5 /*  The type of operating system on which the task
			   is to be created. Currently supports only unix. 
			   Used only by jobmngr to initiate task creation. */ 

#define SIGOFFSET 5    /*  Signals sent from the job manager to the tasks.
			   These are directed to the nodemngrs who propagate
			   them to the tasks */
 
#define SNDR_TID_OFF 5 /*  Tid of sender is put here */

#define REPOFFSET 5    /*  For generic replies */

#define DLEN_OFFSET 6    /*  For task initiation, job manager inserts the 
			     length of the directory name in which the 
			     executable resides, at this offset. */ 

#define SCODEOFFSET 6  /* status code and status messages (reply msgs only) */

#define PNUMOFFSET 6   /*  One byte for port_id. Used by comm routines and 
			   nodemngrs when requesting port addresses of tasks 
			   on remote nodes. */

#define AUTHOFFSET 6   /*  Job mngr puts his sysmngr-assigned authorization 
			   code here, when communicating with nodemngrs */

#define ESTATOFFSET 6  /*  Exit/Error status. Used by job and node mngrs  */

#define SIGSTATOFFSET 7  /* If child died due to a signal, the signal number 
			    is put here */

#define SMSGOFFSET  7  /*  Used for a variety of things, usually for internet
			   addresses (?)  */

#define FLEN_OFFSET 8   /*  For task initiation, job manager inserts the length
			    of the executable file name, at this offset. */ 

#define FDOFFSET    9   /* File descriptor in file operations */

#define  TAG_OFFSET  9 /*  Offset for message tag */


#define NTOFFSET 10    /*  Used for number of tasks, among other things */

#define NUMNODEOFFSET 11 /*  sysmgrs puts the number of allocated nodes here */

#define MSGLEN_OFFSET 14 /* The length of the msg sent by a user-task */

#define NAMEOFFSET 14   /* Offset at which directory and file names are written
			   in create messages */

#define PRM_APPL_MSG_OFFSET 18  /*  Messages sent by appln-tasks start at this
				    offset. Implies that user tasks cannot
				    directly create other tasks or send signals
				    to them. They have to go through the job
				    manager */

#define ERRMSGOFFSET 31  /* If child exits due to an error, nodemngr puts 
			    an error message here */


/* OS codes */

#define UNIX 0  
#define MACH 1

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif

/* Some statically predefined UDP-port numbers */ 

#ifndef NODEMNGR_PORT
#   define NODEMNGR_PORT ARDP_DEFAULT_PORT
#endif

#define SYSMNGR_PORT (NODEMNGR_PORT+1)


/* Task status maintained by nodemngr */

#define T_RUNNING 1
#define T_KILLED  2
#define T_DEAD    3 
#define T_STOPPED 4

/* Status codes for nodes, maintained by sysmngr and nodemngr */

#define NODE_UNUSED    0 /* Node is free; No users or tasks running. */
#define NODE_USER_ON   1 /* user is logged into this workstation node */
#define NODE_JOB_ON    2 /* nodemngr is running tasks on behalf of a jobmngr */
#define NODE_UNAVAIL   3 /* Unknown status; used when sysmngr does not recv
			    reply from nodemngr */

/* Exit codes for jobmngr */

#define J_NORM_EXIT 0  /* Job completed normal execution */
#define J_NO_SM     1  /* Jobmngr couldn't establish connection with sysmngr */
#define J_INSF_RSC  2  /* Job aborted due to insufficient resources */
#define J_NO_LOCIO  3  /* Could not create an IOtask on local node */ 
#define J_NO_NM     4  /* Could not establish connection with nodemngr */
#define J_TOO_MANY  5  /* Too many tasks requested in job description */
#define J_KBD_SIG   6  /* Terminated due to signal from keyboard */
#define J_AUTH_FAIL 7  /* Authorization on node manager failed */
#define J_ARDP_ERR  8  /* Bug in ardp library */
#define J_HOST_TYPE 9  /* Binary type of a task does not match any of the
			  known host types */
#define J_TASK_INIT 10 /* Could not initialize task on node */
#define J_APPL_ERR  11 /* Error occured at application level (term_iotask) */ 

/* status codes for reply messages and return values */

#define SUCCESS 0 
#define FAILURE 1
#define ERRORCODE -1

#define LONG_SZ sizeof(long)

#define NPORTS 4   /* default number of ports per task; */

#define gettid() ((initdone && (_my_taskid >= 0) ) ? _my_taskid : ERRORCODE)

#define getjid() ((initdone && (_my_jobid  >  0) ) ? _my_jobid  : ERRORCODE)


/*
 * getwd: mgrs/iot_create.c mgrs/unix_create.c
 */
#if defined(HPUX)
#define getwd(d)	getcwd(d, MAXPATHLEN)
#define getrusage       hp_getrusage
#endif


/* Support for Heterogeneity */

#define MAX_HOST_TYPES  3  /* Sun3, Sparc, and HPrisc. Add more after we
			      test on other platforms */
#define SUN3   0
#define SUN4   1
#define HP700  2

#define HOSTTYPE(hostname) ((strcasecmp(hostname,"sun3") == 0) ? 0 : \
			    ((strcasecmp(hostname,"sun4") == 0) ? 1 : \
			     ((strcasecmp(hostname,"hp700") == 0) ? 2 : -1)))

#undef MULTI_HOST  /* Do not modify. This is for future use. */

#define PRM_APPL_MSG_LEN 1030
/*** NOTE: This is equal to ARDP_PTXT_LEN - PRM_APPL_MSG_OFFSET - 1. 
 *** The value above should be modified if either of the above two parameters
 *** changes. ARDP_PTXT_LEN is defined in ardp.h and PRM_APPL_MSG_OFFSET is
 *** defined in this file above.
 ***/

typedef struct sockaddr_in  PRM_node_addr;
typedef struct sockaddr_in *PRM_node_addr_t;

#define PRM_AD_SZ sizeof(PRM_node_addr)
#define ADDR_COMPLETE(addr) (addr->sin_port && addr->sin_addr.s_addr)
#define ADDR_REMOTE(tgt_addr,my_addr) \
                     (tgt_addr->sin_addr.s_addr != (my_addr)->sin_addr.s_addr)

#define SEARCH_LIST(head,item,key_field,key_val) \
     item = head; while(item && (item->key_field != (key_val))) \
                      item = item->next


#endif  /* _PRM_H_INCLUDED */

