/*
 * Copyright (c) 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the files
 * <prm-copyr.h>.
 *
 * Written by srao 2/93
 */

#include <prm-copyr.h>

#include <stdio.h>

#include <comm.h>

#include <sys/time.h>

#if defined(HPUX)
#   include <sys/times.h>
#endif


/* ardp_get_dest_addr: Returns the udp socket address (hostaddr + port number),
   given a destination taskid. First checks in the local cache to determine
   whether such a translation has been cached. If so, returns a pointer to
   the cached socket addr. In this case pending_req_ptr will be NOREQ.

   If the address (or part of it) is not in the cache, an an asynchronous 
   translation request is sent to the nodemngr and a pending_req_ptr is set
   to point to this pending reqest. The calling function should later 
   complete the pending request by calling complete_pending_addr_req(), which
   will wait for the request to complete and also cache the translation.
*/

struct sockaddr_in *
  ardp_get_dest_addr(dest_task_id, dest_port_id, pending_reqptr)
int dest_task_id;
u_char dest_port_id;
RREQ *pending_reqptr;
{
  u_short portnum;
  struct sockaddr_in *dest_addrp;
  extern RREQ translate_t2h();
  extern void cache_tid_addr_trsln();

# define CacheTag    trsln_cache[dest_task_id].tag
# define CacheEntry  (trsln_cache[dest_task_id].tinfo)

  if (CacheTag &&
      (dest_port_id < CacheEntry->nports) &&
      ((portnum = *(CacheEntry->ports + dest_port_id)) != 0)) {

    /* translation has already been cached, retrieve cached info */

    dest_addrp = CacheEntry->host_addrp;
    dest_addrp->sin_port = portnum;
    *pending_reqptr = NOREQ;
  }
  else  {
    dest_addrp = (struct sockaddr_in *) malloc(S_AD_SZ);
    bzero(dest_addrp, S_AD_SZ);
    while((*(pending_reqptr) = translate_t2h(dest_task_id, dest_port_id, 
					     dest_addrp, 0))
	  == (RREQ)ERRORCODE) ;
      
    if (*pending_reqptr == NOREQ)
      cache_tid_addr_trsln(dest_task_id, dest_port_id, dest_addrp);
  }
  return dest_addrp;
} 


void
cache_tid_addr_trsln(int dest_task_id, u_char portid, 
		     struct sockaddr_in *in_port_addr)
{
  /* cache the address locally */

  CacheEntry = (struct tasklist *) malloc (sizeof(struct tasklist));
  CacheEntry->host_addrp = in_port_addr;
  CacheEntry->nports = (u_short) ((portid >= NPORTS)? portid + 1: NPORTS);
  CacheEntry->ports = (u_short *)calloc(CacheEntry->nports, sizeof(u_short));
  *(CacheEntry->ports + portid) = in_port_addr->sin_port;

  CacheTag = TRUE;
}


void
complete_pending_addr_req(u_long tid, u_char portid, struct sockaddr_in 
			  *addr, RREQ req)
{
  PTEXT rpkt;
  RREQ newreq;
  extern RREQ process_trsln_req(), translate_t2h();

  newreq = req;
  if (ardp_retrieve(newreq, -1) == ARDP_SUCCESS) {
    newreq = process_trsln_req(req, addr);
    while ( newreq == (RREQ)ERRORCODE)
      newreq = translate_t2h(tid, portid, addr, -1);

    cache_tid_addr_trsln(tid, portid, addr);

  }
}


RREQ
translate_t2h(remote_task_id, port_id, destp, wait_time)
int remote_task_id, wait_time;
u_char port_id;
struct sockaddr_in *destp;
{
  u_char op;
  PTEXT pkt, rpkt;
  RREQ newreq, prcd_req;
  char *msg;
  u_long ntmp1;
  int send_status;
  extern RREQ process_trsln_req();

  op = remote_task_id ? HADD_REQ: IJADD_REQ;
  pkt = ardp_ptalloc();
  msg =  pkt->start;
  *(msg + PRM_OP_OFFSET) = op;
  
  ntmp1 = htonl(_my_jobid);
  bcopy(&ntmp1, msg + PRM_JIDOFFSET, LONG_SZ);
  ntmp1 = htonl(remote_task_id);
  bcopy(&ntmp1, msg + NTOFFSET, LONG_SZ);
  *(msg + PNUMOFFSET) = port_id;
  pkt->length = NTOFFSET + LONG_SZ;
  
  newreq = ardp_rqalloc();
  newreq->outpkt = pkt;
  
  if ((send_status = ardp_send(newreq, _my_hostname, 0, wait_time) ) 
       == ARDP_PENDING)
    return newreq;
  
  if (send_status == ARDP_SUCCESS) {
    prcd_req = process_trsln_req (newreq, destp);
    ardp_rqfree(newreq);
    return prcd_req;
  }
}


RREQ
  process_trsln_req(RREQ req, struct sockaddr_in *addrp)
{
  PTEXT rpkt;
  char *msg;
  
  if( (rpkt = req->rcvd) == NOPKT) {
    if (prm_debug) 
      fprintf(stderr, "(jid=%lu tid=%d) Address request to nodemngr %s timed out.\n",
	      _my_jobid, _my_taskid, _my_hostname);
    return (RREQ)ERRORCODE;
  }
  
  msg = rpkt->start;
  if ( (*(msg + REPOFFSET)== HOSTID) && (*(msg + SCODEOFFSET) == SUCCESS)) 
    bcopy(msg + SMSGOFFSET, addrp, S_AD_SZ);
  else {
    if(prm_debug)
      fprintf(stderr, "(jid=%lu tid=%d) Nodemngr could not perform t -> h translation!\n", _my_jobid, _my_taskid);
    return (RREQ)ERRORCODE;
  }
  if(addrp->sin_port)
    return NOREQ;
  else
    return (RREQ)ERRORCODE;
}

