/*
 * Copyright (c) 1992, 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the files
 * <prm-copyr.h>.
 *
 * Written by srao 12/92
 *
 */

#include <prm-copyr.h>

#include <comm.h>

#ifdef DEBUG
#   ifdef HPUX
#      include <signal.h>
#   endif
#endif

vsend_v(dest_task_id, dest_port_id, tag, dataptr, elem_len, stride, elem_cnt)
int dest_task_id, tag;
u_char dest_port_id;
char *dataptr;
int elem_len, stride, elem_cnt;
{
  PTEXT pkt_0, pkt_i, prev_pkt;
  RREQ newreq;
  u_long ntmp1, i, byte_dist, data_len;
  int pkt_num, elems_per_pkt, npkts, reply_code;
  char *start, *from, *to;
  PRM_node_addr_t dest_host_addr;
  extern PRM_node_addr_t get_dest_addr();

  /* Obtain the address of the host on which destination task is running.
     This could be from the local cache of the process, local nodemngr's 
     cache or the remote nodemngr's cache or the jobmngr */
  
  dest_host_addr = get_dest_addr(dest_task_id, dest_port_id);
  
  if (dest_task_id == _my_taskid) {
    strcpy(p_err_string, ": Cannot send message to myself.");
    return ERRORCODE;
  }
  
  data_len = elem_len * elem_cnt;

  pkt_0 = ardp_ptalloc();
  start = pkt_0->start;
  *(start + PRM_OP_OFFSET) =  (u_char)APP_MSG;
  ntmp1 = htonl(_my_taskid);
  bcopy(&ntmp1, start + SNDR_TID_OFF, LONG_SZ);
  ntmp1 = htonl(tag);
  bcopy (&ntmp1, start + TAG_OFFSET, LONG_SZ);
  
  from = dataptr;
  to = start + PRM_APPL_MSG_OFFSET;
  pkt_i = pkt_0;

  npkts = data_len/PRM_APPL_MSG_LEN + (data_len % PRM_APPL_MSG_LEN ? 1 : 0);
  if (npkts < 2) {
    for (i = 0; i < elem_cnt; i++ ) {
      bcopy(from, to, elem_len);
      from += stride;  /* Note that stride is specified in bytes, 
			  not words */
      to += elem_len;
    }    
    pkt_0->length = data_len + PRM_APPL_MSG_OFFSET;
  }
  else {
    /* fprintf(stderr, "task %d -- multipacket send\n", _my_taskid); */
    elems_per_pkt = elem_cnt/npkts + (elem_cnt % npkts ? 1 : 0) ;
    for (i = 0; i < elems_per_pkt; i++ ) {
      bcopy(from, to, elem_len);
      from += stride;  /* Note that stride is specified in bytes, 
			  not words */
      to += elem_len;
    }    
    pkt_0->length = PRM_APPL_MSG_OFFSET + PRM_APPL_MSG_LEN;
    prev_pkt = pkt_0;
    for (pkt_num = 1; pkt_num < npkts; pkt_num++) {
      pkt_i = ardp_ptalloc();
      to = pkt_i->start + PRM_APPL_MSG_OFFSET;
      for (i = 0; i < elems_per_pkt; i++ ) {
	bcopy(from, to, elem_len);
	from += stride;  /* Note that stride is specified in bytes,
			       not words */
	to += elem_len;
      }    
      pkt_i->length = PRM_APPL_MSG_OFFSET + PRM_APPL_MSG_LEN;
      prev_pkt->next  = pkt_i;
      pkt_i->previous = prev_pkt;
      prev_pkt = pkt_i;
    }
    pkt_0->previous = pkt_i;
  }

  newreq = ardp_rqalloc();
  newreq->outpkt = pkt_0;
  if ((perrno = ardp_send(newreq, 0, dest_host_addr, -1)) != ARDP_SUCCESS) {
    sprintf(p_err_string, ": ardp_send to %d returned %d", dest_task_id, 
	    perrno);
#ifdef DEBUG
    io_printf("vsend_v%s",  &p_err_string, (char *)0);
    if (_my_taskid)
      kill (getpid(), SIGSTOP);
#endif    
    return ERRORCODE;
  }

  if(newreq->rcvd)              /* reply recvd */
    reply_code = (int)*(newreq->rcvd->start + PRM_OP_OFFSET);
  else {
    reply_code = ERRORCODE;
    sprintf (p_err_string, ": Could not reach task %d", dest_task_id);
  }
  ardp_rqfree(newreq);
  return reply_code;
}



vrecv_v(sndr_task_id, rcv_port_id, tag, rcv_buffer, elem_len, stride, elem_cnt)
int sndr_task_id, tag;
u_char rcv_port_id;
char *rcv_buffer;
int elem_len, stride, elem_cnt;
{
  RREQ msgreq;
  PTEXT pkt_0, pkt_i, rpkt;
  int i, npkts, curlen; 
  u_long rcvd_len;
  u_long byte_dist, xpctd_len, copied_len ;
  int to_copy_len, cnt, nbytes;
  char *start, *from, *to;

  if (sndr_task_id == _my_taskid) {
    sprintf(p_err_string, ": Cannot receive message from myself.");
    return ERRORCODE;
  }

  msgreq = get_message(sndr_task_id, tag);
  pkt_0 = msgreq->rcvd;
  start = pkt_0->start;
  
  xpctd_len = elem_len * elem_cnt;

  if( *(start + PRM_OP_OFFSET) == APP_MSG){
    if ( (xpctd_len == 0) || (rcv_buffer == NULL) ) { 
      copied_len = 0;
      goto reply;
    }
    bcopy (start + MSGLEN_OFFSET, &rcvd_len, LONG_SZ);
    rcvd_len = ntohl(rcvd_len);

    to_copy_len = MIN(rcvd_len, xpctd_len);

    if( to_copy_len <= PRM_APPL_MSG_LEN) {   /* single packet message */
      cnt = to_copy_len/elem_len;
      scatter_data(start + PRM_APPL_MSG_OFFSET, rcv_buffer, elem_len, stride, cnt);
    }
    else {  /* multipacket message */
      copied_len = 0;
      pkt_i = pkt_0;
      cnt = PRM_APPL_MSG_LEN/elem_len;
      while (pkt_i) {
	if ( (nbytes = to_copy_len - copied_len) >  PRM_APPL_MSG_LEN) {
	  scatter_data(pkt_i->start + PRM_APPL_MSG_OFFSET, rcv_buffer,
		  elem_len, stride, cnt);
	  copied_len += PRM_APPL_MSG_LEN;
	  rcv_buffer += PRM_APPL_MSG_LEN;
	}
	else {
	  if (nbytes > 0) {
	    cnt = nbytes/elem_len;
	    scatter_data(pkt_i->start + PRM_APPL_MSG_OFFSET, rcv_buffer, 
		    elem_len, stride, cnt);
	    copied_len += nbytes;
	  }
	  break;
	}
	pkt_i = pkt_i->next;
      }
    }

  reply:
    msgreq->outpkt = ardp_ptalloc();
    msgreq->outpkt->seq = 0;
    *(msgreq->outpkt->start + PRM_OP_OFFSET) = SUCCESS;
    msgreq->outpkt->length = SCODEOFFSET + 1;
    ardp_respond(msgreq, ARDP_R_COMPLETE);   /* Send reply to sender task */  
  }


}


void
vsendrecv(src_task_id,   src_tag, inbufp,   inlen, 
         dest_task_id, dest_tag, outbufp, outlen)

int src_task_id, src_tag, inlen, dest_task_id, dest_tag, outlen;
char *inbufp, *outbufp;
{
#define Dest_Port_Id 1
#define Src_Port_Id  1

  if (src_task_id < _my_taskid) {
    while (vsend (dest_task_id, Dest_Port_Id, dest_tag, outbufp, outlen)
	   != SUCCESS);
    vrecv (src_task_id, Src_Port_Id, src_tag, inbufp, inlen);
  }
  else {
    vrecv (src_task_id, Src_Port_Id, src_tag, inbufp, inlen);
    while(vsend (dest_task_id, Dest_Port_Id, dest_tag, outbufp, outlen)
	  == ERRORCODE)
      io_printf("vsend in vsendrecv%s", p_err_string, (char *)0);;   
  }
}


scatter_data(char *start, char *dest, int elem_len, int stride, int cnt)
{
  char *until_ptr; 
  char *from = start;
  char *to = dest;

  until_ptr = start + elem_len * cnt;
  while (from < until_ptr) {
    bcopy(from, to, elem_len);
    to += stride;
    from += elem_len;
  }
}


reduce_int(value, redn_op)
int value, redn_op;
{
  int ret_val;
  vsend(0, (u_char)1, redn_op, (char *)&value, sizeof(int));
  vrecv(0, (u_char)1, redn_op, (char *)&ret_val, sizeof(int));
  return ret_val;
}

