/*
 * Copyright (c) 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the files
 * <prm-copyr.h>.
 *
 * Written by srao 3/93
 *
 */

#include <prm-copyr.h>


#include <stdio.h>

#ifndef MACH386
#include <stdlib.h>
#endif

#ifdef HPUX
#   include <signal.h>
#endif

#if defined(SUNOS)
#   include <vfork.h>
#endif

#include <sys/time.h>
#include <sgtty.h>   
#include <fcntl.h>

#define PRM_PROG
#include <comm.h>

#include <jobmngr.h>

extern TDES *ttbl;
extern u_long ntasks_running;
extern PRM_node_addr   my_node_addr;

process_init_req(RREQ req, int taskid)
{
  int st;
  u_short udp_port, emsglen;
  char *msg;

  st = process_pending_req(req);
  if ( st == J_NO_NM) 
    return st;

  msg = req->rcvd->start;
  if (st == J_AUTH_FAIL) {
    bcopy(msg + ERRMSGOFFSET, &emsglen, 2);
    emsglen = ntohs(emsglen);
    if (emsglen)
      bcopy(msg + ERRMSGOFFSET + 2, p_err_string, emsglen);
    
    return J_TASK_INIT;
  }
  bcopy(msg + SMSGOFFSET, &udp_port, 2);
  ttbl[taskid]->port = (u_short)udp_port;    /* In network_byte order */ 
  ++ntasks_running;
  ttbl[taskid]->status = T_RUNNING;
  return SUCCESS;
}


process_pending_req(RREQ req)
{
  u_char reply_op;
  PTEXT rpkt;
  RREQ cmpreq;
  extern RREQ ardp_completeQ;

  if (req->status != ARDP_STATUS_COMPLETE) {
    if(ardp_retrieve(req, -1) != ARDP_SUCCESS)
      return J_ARDP_ERR;
  }
  else {
    cmpreq = ardp_completeQ;
    while(cmpreq && ( cmpreq != req)) 
      cmpreq = cmpreq->next;
    if (cmpreq)
      EXTRACT_ITEM(req, ardp_completeQ);
  }
  if( (rpkt = req->rcvd) == NOPKT) {  /* No ack recvd */
    fprintf(stderr, "(%s) nodemngr at %s not responding!!!\n",
	    _progname, inet_ntoa(req->peer_addr) );
    return J_NO_NM;
  }
  else {  /* Check if authorization was successful  */
    reply_op = *(rpkt->start + PRM_OP_OFFSET);
    if (reply_op == FAILURE)  {
      return J_AUTH_FAIL;
    }
    else
      return 0;
  }
}


/* PRM_acquired_resources : request resource allocation from sysmngr(s) 
   and acquire them. 
   Sysmngr host names have been read from the job desc. file into 
   sm_hostlist. Try to acquire nodes from them, in the order specified.
   NOTE that these are strictly 'alternate' sysmngrs in that the jobmngr
   cannot acquire a few resources from one and the rest from another. If
   one cannot allocate sufficent resources, it just ignores that sysmngr and
   goes to the next one. 
*/


extern int num_sm_hosts, ntasks;
extern char *sm_hostname, *sm_hostlist[], buffer[];

PRM_acquire_resources()
{
  int i, rsc_acqd, ret_code;
  RREQ newreq;
  PTEXT pkt, rpkt;
  char *msg;
  u_long njid, ntmp;

  rsc_acqd = FALSE;
  i = 0;
  
  while (!rsc_acqd && (i < num_sm_hosts)  ) {
    sm_hostname = sm_hostlist[i++];
    sprintf(buffer,"%s(%d)", sm_hostname, SYSMNGR_PORT);
    
    if(prm_debug)
      fprintf(stderr, "(%s) Requesting %d nodes from sysmngr %s .....\n", 
	      _progname, ntasks, sm_hostname);
    
    pkt = ardp_ptalloc();
    msg = pkt->start;
    njid = htonl(_my_jobid);
    bcopy(&njid, msg + PRM_JIDOFFSET, LONG_SZ);
    *(msg + PRM_OP_OFFSET) = RSC_ALLOC;
    pkt->length = NTOFFSET;
    
    
#if MULTI_HOST
    for (i = 0; i < MAX_HOST_TYPES; i++ ) {
      ntmp = htonl(h_type_reqd[i]);
      bcopy(&ntmp, msg + pkt->length, LONG_SZ);
      pkt->length += LONG_SZ;
    }
#else
    ntmp = htonl(ntasks);
    bcopy(&ntmp, msg + pkt->length, LONG_SZ);
    pkt->length += LONG_SZ;
#endif
    
    newreq = ardp_rqalloc();
    newreq->outpkt = pkt;
    ardp_send (newreq, buffer, 0, -1);
    
    if ((rpkt = newreq->rcvd) == NOPKT) {
      fprintf(stderr, "(%s) No response from sysmngr %s.\n", _progname, 
	      sm_hostname);
      ardp_rqfree(newreq);
      ret_code = J_NO_SM;
    }
    else {
      msg = rpkt->start;
      if (*(msg + REPOFFSET) == FAILURE) {
	fprintf (stderr, "(%s) sysmngr %s does not have free nodes.\n",
		 _progname, sm_hostname);
	ardp_rqfree(newreq);
	ret_code = J_INSF_RSC;
      }
      else
	rsc_acqd = TRUE;
    }
  } /* while */
  return ret_code;
}


PTEXT
  PRM_request_resources()
{
  PTEXT pkt, rpkt;
  RREQ newreq;
  char *msg;
  u_long ntmp;

  pkt = ardp_ptalloc();
  msg = pkt->start;
  ntmp = htonl(_my_jobid);
  bcopy(&ntmp, msg + PRM_JIDOFFSET, LONG_SZ);
  *(msg + PRM_OP_OFFSET) = RSC_ALLOC;
  pkt->length = NTOFFSET;
  
#if MULTI_HOST
  for (i = 0; i < MAX_HOST_TYPES; i++ ) {
    ntmp = htonl(h_type_reqd[i]);
    bcopy(&ntmp, msg + pkt->length, LONG_SZ);
    pkt->length += LONG_SZ;
  }
#else
  ntmp = htonl(ntasks);
  bcopy(&ntmp, msg + pkt->length, LONG_SZ);
  pkt->length += LONG_SZ;
#endif
  
  newreq = ardp_rqalloc();
  newreq->outpkt = pkt;
  if ((perrno = ardp_send (newreq, buffer, 0, -1)) != ARDP_SUCCESS)
    rpkt = NOPKT;
  else {
    rpkt = newreq->rcvd;
    newreq->rcvd = NOPKT;
    ardp_rqfree(newreq);
  }
  return rpkt;
}



JM_send_tmap_to_nm(addr, map_pkt)
PRM_node_addr_t addr;
PTEXT map_pkt;
{
  int retval;
  PTEXT pkt, rpkt;
  RREQ newreq;
  u_char reply_op;

  retval = SUCCESS;

  pkt = ardp_ptalloc();
  bcopy(map_pkt->start, pkt->start, map_pkt->length);
  pkt->length = map_pkt->length;
  
  addr->sin_port = htons((u_short) NODEMNGR_PORT);
  
  newreq = ardp_rqalloc();
  newreq->outpkt = pkt;
  ardp_send (newreq, 0, addr, -1);
  
  if ( (rpkt = newreq->rcvd) == NOPKT) {
    if (ADDR_REMOTE(addr,&my_node_addr))
      fprintf(stderr, "(%s) nodemngr %s is not responding\n", _progname, 
	      inet_ntoa(addr->sin_addr));
    else
      fprintf(stderr, "(%s) Local nodemngr is not responding\n", _progname); 
    retval = J_NO_NM;
  }
  else {  /* Check if authorization was successful  */
    reply_op = *(rpkt->start + REPOFFSET);
    if (reply_op == FAILURE) {
      if (ADDR_REMOTE(addr,&my_node_addr))
	fprintf(stderr, "(%s) Authorization on nodemngr %s failed!\n", 
		_progname, inet_ntoa(addr->sin_addr)); 
      else
	fprintf(stderr, "(%s) Authorization on local nodemngr failed!\n", 
		_progname); 
      retval = J_AUTH_FAIL;
    }
  }    
  ardp_rqfree(newreq);
  return retval;
}
