/*
 * Copyright (c) 1992, 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the files
 * <prm-copyr.h>.
 *
 * Modified by srao 6/92 - 3/93
 *
 */

#include <prm-copyr.h>

#include <stdio.h>

#include <signal.h>       /* To handle signals from child tasks, timers */
#include <sys/wait.h>     /* To obtain exit status of child */
#include <utmp.h>
#include <sys/utsname.h>

#define  MAIN_PROG
#include <comm.h>

#include <nodemngr.h>

char    *progname, *loc_bin_dir;
char    my_hostname[MAXHOSTNAMELEN], *smhostname, *node_archname;
char    buf[128];

int     utflag = FALSE;       /* Set if timer is enabled for utmp checking */
int     avail_win_flag = FALSE; /* Set if nodemngr is available only within
				   specified time windows */
int     win_start, win_end;
int     win_start_mins, win_end_mins;
int     node_archtype;

u_char  node_status=NODE_UNUSED;  /* See nodemngr.h for desc. of statuses */
u_char  tmp_status;

u_short my_udp_portnum, next_udp_port = FIRST_USR_PORT; 

long    timer_intvl[4];       /* Interval in secs, for timer alarm 
				 when the workstation is free/busy */
u_long  inaddr;

PRM_node_addr   my_node_addr;     /* addr of nodemngr host */
PRM_node_addr   h_ad, smaddr;  /* For other addresses */
PRM_node_addr_t dest, tmpaddr;   

struct  pidlist *pid_table=NULL,  /* Contains mapping of unix pid to 
					  task-information */
                *cur_pid_ent;

struct  timeval task_start_time;
struct  itimerval timeval, 
                  otimeval;   /* To handle timers , interrupts */

struct  authlist *authl_head=NULL, *cur_al;
struct  hostent *hent;

JINF    job_tbl = NULL;    /* Job table is implemented as a linked list of
			      structures, each node in list holds info about
			      a job whose task(s) is running on this node.*/

extern  int unix_create_task();
extern  void NM_sigchld_hndlr();     /* Signal handler for SIGCHLD */
extern  void NM_update_node_status();      /* Signal handler for SIGALRM */


int     (*os_create[])()= {  unix_create_task   };


main(int argc, char *argv[], char *envp[])
{

  struct utsname unamebuf;

  timer_intvl[NODE_UNUSED] = 60; 
  timer_intvl[NODE_USER_ON] = 900;
  timer_intvl[NODE_JOB_ON] = 0;
  win_start = 0000;
  win_end = 0600;
  bzero(&smaddr, PRM_AD_SZ);
  loc_bin_dir = LOC_BIN_DIR;

  _progname = progname = argv[0];
  argc--;argv++;
  
  while (argc > 0 && **argv == '-') {
    switch (*(argv[0]+1)) {
    case 'D':
            prm_debug=1;    /* Default debug level */
	    sscanf(argv[0],"-D%d", &prm_debug);
            break;

    case 'i':
	    utflag = UT_SET;
	    sscanf(argv[0],"-i%d:%d", &timer_intvl[NODE_UNUSED], 
		   &timer_intvl[NODE_USER_ON]);
	    if(timer_intvl[NODE_UNUSED] <= 0) timer_intvl[NODE_UNUSED] = 60; 
	    if(timer_intvl[NODE_USER_ON] <= 0) timer_intvl[NODE_USER_ON] = 900;
	    break;

   case 'l':
	    argc--; argv++;
	    loc_bin_dir = (char *)malloc(strlen(argv[0]) + 1);
	    sscanf(argv[0], "%s", loc_bin_dir);
	    break;

   	    
   case 's':
	    argc--; argv++;
	    smhostname = (char *)malloc(strlen(argv[0]) + 1);
	    sscanf(argv[0], "%s", smhostname);
	    break;

   case 't':
	    avail_win_flag = TRUE;
	    sscanf(argv[0],"-t%d:%d", &win_start, &win_end);
	    win_start_mins = (win_start/100) * 60 + (win_start%100);
	    win_end_mins  = (win_end/100) * 60 + (win_end%100);
	    break;
	    

    default:
            fprintf(stderr, "Usage: %s [ -D[debug-level] ] [ -i[intvl_unused_secs:intvl_user_secs] ]\n [ -s sysmngr-hostname ] [ -twindow_start:window_end ]\n", 
		    progname);
            exit(1);
	  }
    argc--; argv++;
  }

  pfs_debug = prm_debug;

#ifndef HPUX  /* setreuid not available on HPUX */

/* This will not be necessary in the long run. provided now for those cases 
   when users start their own nodemngrs */

  if (setreuid(geteuid(), -1) == -1) {
    fprintf(stderr, "could not set uid:\n");
    perror("");
  }

#endif

  /* We have to set the timer to periodically check utmp file and/or 
     check whether we are within the permissible time window */
  
  if( (utflag == UT_SET) ||  (avail_win_flag) ) {
    
    update_node_status();
    
    /* Set next expiry time to the smallest of: seconds given on command line,
       time-to-start of window boundary, time-to-end of window boundary  */
    
    timeval.it_value.tv_sec = timeval.it_interval.tv_sec = 
      smallest_time_to_expiry(timer_intvl[node_status], win_start, win_end);
    
    timeval.it_value.tv_usec = timeval.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &timeval, &otimeval);
    signal(SIGALRM, NM_update_node_status );
    utflag |= TM_ENBL;
  }
  
  gethostname(my_hostname, sizeof(my_hostname));
  hent = gethostbyname(my_hostname);
  bcopy(hent->h_addr, &inaddr, 4);
  bzero((char *)&my_node_addr, PRM_AD_SZ);
  my_node_addr.sin_family = hent->h_addrtype;
  bcopy(hent->h_addr, (char *)&(my_node_addr.sin_addr), hent->h_length);
  my_node_addr.sin_port = htons((u_short) NODEMNGR_PORT);
  
  sprintf(buf,"#%d", NODEMNGR_PORT);
  my_udp_portnum = ardp_bind_port(buf);
  
  signal(SIGCHLD, NM_sigchld_hndlr);


  if (uname(&unamebuf) == -1) {
    (void)printf("(%s) Could not get host architecture type\n", progname);
    exit(1);
  }
  else
    node_archname = ((strncmp(unamebuf.machine, "sun4", 4) == 0) ? "sun4" :
		     ((strncmp(unamebuf.machine, "9000/", 5) == 0) ? "hp700" :
		      "unknown"));

  node_archtype = HOSTTYPE(node_archname);

  /* NODEMNGR'S LOOP  */ 

  for (;;) {   

    NM_Process_Req();

  } 

}  



/* Nodemngr's main procedure: waits until a  request arrives at the port and 
   processes it. The processing performed depends on the PRM_opcode in the 
   request. */


NM_Process_Req()
{
  PTEXT  pkt, rpkt;
  RREQ    current_req, newreq;
  

  int     child_pid;
  int     i, timetowait;
  int     jobfound, taskfound, num_task_ent, arc_stat;
  
  char    filename[100], dirname[MAX_DIR_LINESIZE];
  char    *msg, *curplace; /* used as generic position pointers in message 
			      buffers */
  char    *tbits;          /* Cache tag */
  u_char  osname;
  u_char  req_op, req_pnum, signum;
  u_short filelength, dirlength, ndeg, np;
  u_short emsglen, nelen;   /* length of error messages */
  u_long  req_tid, curtid; 
  u_long  req_jid, req_auth;
  u_long  ntmp1;
  JINF    curjob;
  static  struct  taskinfo *curtask;

  current_req = ardp_get_nxt();
  msg = current_req->rcvd->start;
  
  /* Every message has an operation code giving an indication of what is
     to be done */
  req_op = *(msg + PRM_OP_OFFSET);
  
  /* Get the jobid of the sender from the message. Valid only for jobmngr 
     and tasks. For other entities, we don't care what it is. */ 
  bcopy(msg + PRM_JIDOFFSET, &req_jid, LONG_SZ);  
  req_jid = ntohl(req_jid);
  
  switch (req_op) {
    
  case TASK_INIT: /* request from a jobmngr for task creation. nodemngr
		     uses the pathname, taskid  and jobid specified by the 
		     jobmngr to fork a child process. On success, pid of
		     task is returned, and job/task info is updated. */
    
    osname = *(msg + OSCODE_OFFSET);
    bcopy(msg + DLEN_OFFSET, &dirlength, 2);  /* len includes null 
						 terminator */
    dirlength = ntohs(dirlength);
    if (dirlength) {
      bcopy(msg + NAMEOFFSET, dirname, dirlength);
      sprintf(dirname + dirlength - 1,"/%s", node_archname);
    }
    else
      dirname[0]='\0';
    
    bcopy(msg + FLEN_OFFSET, &filelength, 2);
    filelength = ntohs(filelength);
    bcopy(msg + NAMEOFFSET + dirlength, filename, filelength);
    
    bcopy(msg + NTOFFSET, &req_tid, LONG_SZ);
    req_tid = ntohl(req_tid);
    
    /* Look up job table, ensure that there is an entry for this job */
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);
    if (curjob == NULL) { 
      sprintf(p_err_string, "Job not authorized on node %s",
	      inet_ntoa(inaddr));
      
      prm_server_respond(current_req, (u_char)FAILURE, (u_char)FAILURE,
			 0, p_err_string);
      NM_inform_sm_node_status(node_status);
      break;
    }
    
    /* Inform jobmngr not to retry request for some time */
    timetowait = ((strncasecmp(dirname, "io:", 3) == 0) ? 80 : 5);
    if (ardp_rwait(current_req, timetowait, 0, 0) != PSUCCESS)
      fprintf(stderr, "(%s) Could not send backoff pkt to jobmngr\n", 
	      progname);
    
    if (utflag == (UT_SET | TM_ENBL)){/* utmp checking set and timer not
					 already disabled. */
      timeval.it_value.tv_sec = timeval.it_interval.tv_sec = 0;
      setitimer(ITIMER_REAL, &timeval, &otimeval);
      utflag ^= (TM_ENBL | DIS_TMP);   /* Disable timer and indicate
					  temporarily disabled */ 
    }
    gettimeofday( &task_start_time, (struct timezone *)0);
    child_pid = (*os_create[osname])(dirname, filename, req_tid, req_jid, 
				     next_udp_port, 0);
    if ( (child_pid != ERRORCODE) && ((emsglen = strlen(p_err_string) + 1) 
				      == 1) )   {
      rpkt = current_req->outpkt = ardp_ptalloc();
      ntmp1 = htonl(child_pid);
      bcopy(&ntmp1, rpkt->start + PRM_JIDOFFSET, 4);
      *(rpkt->start + PRM_OP_OFFSET) = (u_char)TINIT_SCS;
      *(rpkt->start + SCODEOFFSET) = (u_char)SUCCESS;
      nelen = htons(next_udp_port);
      bcopy(&nelen, rpkt->start + SMSGOFFSET, 2);
      rpkt->length = SMSGOFFSET + 2;
      ardp_respond(current_req, ARDP_R_COMPLETE);
    }
    else {     /* Failed to spawn a process */
      prm_server_respond(current_req, (u_char)FAILURE, (u_char)FAILURE,
			 0, p_err_string);
      NM_inform_sm_node_status(node_status);
      
      if (utflag == (UT_SET | DIS_TMP)) { /* Timer was recently disabled. 
					     re-enable it */
	timeval.it_value.tv_sec = timeval.it_interval.tv_sec = 
	  smallest_time_to_expiry(timer_intvl[node_status], win_start,
				  win_end);
	setitimer(ITIMER_REAL, &timeval, &otimeval);
	utflag ^= (TM_ENBL |DIS_TMP) ;  
      }
      break;
    }
    if (utflag == (UT_SET | DIS_TMP) )
      utflag ^= DIS_TMP;
    
    if (req_tid == 0) { /* This is the fileio task */
      curjob->fioaddr.sin_port = htons(next_udp_port);
      ++next_udp_port;    /* allocate one port to the term_iotask */
      curjob->iotaddr.sin_port = htons(next_udp_port);
      goto task_update_done;
    }
    /* when nodemngr starts caching task to hostaddr mapping, we need to
       add an entry in the job table for the task just created. */
    
    SEARCH_LIST(curjob->taskptr,curtask,tid,req_tid);
    if (!curtask) {
      curtask = (struct taskinfo *) malloc(sizeof(struct taskinfo));
      curtask->tid = req_tid; /* tasks are numbered 1 through ntasks */
      curjob->cache_tag[i] = 1;
      curtask->ports = (u_short *)calloc(NPORTS, sizeof(u_short));
      curtask->nports = NPORTS;
    }
    curtask->status = T_RUNNING;
    curtask->start_time.tv_sec  = task_start_time.tv_sec;
    curtask->start_time.tv_usec = task_start_time.tv_usec;
    *(curtask->ports + 1) = htons(next_udp_port);
    cur_pid_ent = (struct pidlist *)malloc(sizeof(struct pidlist));
    curtask->pid_tbl_p = cur_pid_ent;
    cur_pid_ent->pid = child_pid;
    cur_pid_ent->taskptr = curtask;
    cur_pid_ent->jobptr = curjob;
    APPEND_ITEM(cur_pid_ent, pid_table);
    
    node_status = NODE_JOB_ON;
    
  task_update_done:
    ++next_udp_port;
    if (next_udp_port == 65535)
      next_udp_port = FIRST_USR_PORT;
    
    break;
    
    
  case TTBL_UPDATE:  /* Job manager has sent the mapping of tasks to nodes.
			Cache it in my local address space */
    
    bcopy (msg + AUTHOFFSET, &req_auth, LONG_SZ);
    
    if (check_auth(req_jid, req_auth, &(current_req->peer) ) == FAILURE) {
      sprintf(p_err_string, "(job %d) Authorization on nodemngr failed.",
	      req_jid);
      prm_server_respond(current_req, (u_char)FAILURE, (u_char)FAILURE,
			 0, p_err_string);
      
      break;
    }
    
    /* Number of task entries to update */
    bcopy(msg + NTOFFSET, &num_task_ent, LONG_SZ); 
    num_task_ent = ntohl(num_task_ent);
    
    /* Check if an entry for this job  already exists */
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);
    if (!curjob) {             /* No entry. This is a new job */
      curjob = (JINF)malloc (sizeof(struct job_info));
      curjob->jid = req_jid;
      curjob->numtasks = 0;
      curjob->taskptr = NULL;
      APPEND_ITEM(curjob, job_tbl);
      
      curjob->cache_tag = (char *)malloc(num_task_ent + 1);
      bzero(curjob->cache_tag, num_task_ent + 1);
      /* Cache tag bits used to check whether an entry for a task exists */
    } /* if (!curjob) */
    
    tbits = curjob->cache_tag;
    
    curplace = msg + NTOFFSET + LONG_SZ;
    
    for (i = 0; i <= num_task_ent; i++) { 	  
      bcopy(curplace, &curtid, LONG_SZ);/* get tid of current task entry */
      curtid = ntohl(curtid);
      curplace += LONG_SZ;
      
      if ( tbits[curtid] )  {
	/* An entry for this task exists  in job table */
	/* look for the entry */
	SEARCH_LIST(curjob->taskptr,curtask,tid,curtid);
      }
      else  {
	curtask = (struct taskinfo *) malloc(sizeof(struct taskinfo));
	curtask->tid = curtid; /* tasks are numbered 1 through ntasks */
	tbits[i] = 1;
	curtask->ports = (u_short *)calloc(NPORTS, sizeof(u_short));
	curtask->nports = NPORTS;
	curtask->pid_tbl_p = NULL;
	
	curtask->next = curjob->taskptr;
	curjob->taskptr = curtask;
	if (curtid != 0)  /* This is not an iotask */
	  ++(curjob->numtasks);
      }
      bcopy(curplace, &(curtask->hostaddr), PRM_AD_SZ);
      curtask->hostaddr.sin_family = AF_INET;
      curplace += PRM_AD_SZ;
    } /* for */
    
    bcopy(&(current_req->peer), &(curjob->jmaddr), PRM_AD_SZ);
    
    /* NOTE: this assumes that jobmngr doesn't migrate, so the host 
       address of IOtasks are the same as that of jobmngr */
    bcopy(&(current_req->peer), &(curjob->iotaddr), PRM_AD_SZ);
    bcopy(&(current_req->peer), &(curjob->fioaddr), PRM_AD_SZ);
    
    curjob->iotaddr.sin_family = curjob->jmaddr.sin_family = AF_INET;
    curjob->fioaddr.sin_port = curjob->iotaddr.sin_port = 
      curjob->jmaddr.sin_port = 0;
    
    /* Send reply to jobmngr */
    prm_server_respond(current_req, (u_char)SUCCESS, (u_char)SUCCESS,
		       0, (char *)0);
    
    break;
    
    
  case PMAP_REQ:  /* A request from remote nodemngr for l->p mapping */
  case HADD_REQ:   /* A request from local task for t->h mapping */
    
    bcopy(msg+NTOFFSET, &req_tid, LONG_SZ);
    req_tid = ntohl(req_tid);
    req_pnum = *(msg+PNUMOFFSET);
    
    /* Inform task not to retry request for another 15 secs */
    if (ardp_rwait(current_req, 15, 0, 0) != PSUCCESS)
      fprintf(stderr, "(%s) Could not send backoff pkt to task.\n", 
	      progname);
    
    jobfound = FALSE;
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);

    if(curjob) {
      register long mask=0x1;
      jobfound =TRUE;
      if((curjob->cache_tag)[req_tid] ){ /* the t->h mapping is cached */
	SEARCH_LIST(curjob->taskptr,curtask,tid,req_tid);
	if (curtask) {
	  bcopy(&(curtask->hostaddr), &h_ad, PRM_AD_SZ);
	  if ( (req_pnum > curtask->nports) || 
	      ((h_ad.sin_port = *(curtask->ports+req_pnum)) == 0)) {
	    
	    if ( (h_ad.sin_addr.s_addr != inaddr) && (req_op 
						      == HADD_REQ)) { 
	      
	      get_l2p_from_remote(req_pnum, req_jid, req_tid, &h_ad);
	      /* cache it locally */
	      if ( req_pnum >= curtask->nports) {
		curtask->nports = req_pnum + 1;
		curtask->ports = (u_short *)realloc(curtask->ports,
						    req_pnum * 2);
	      } 
	      *(curtask->ports + req_pnum) = h_ad.sin_port;
	    }
	    else   /* the requested l->p mapping does not exist */
	      h_ad.sin_port = 0;
	  }
	}
	else
	  h_ad.sin_port = 0;
      }
      else  /* get t->h mapping from jm and l->p mapping from nm */
	(curjob->cache_tag)[req_tid] = 1;
    }
    rpkt = current_req->outpkt = ardp_ptalloc();
    rpkt->seq = 0;
    msg = rpkt->start;
    
    if ( (curjob == NULL) ){
      *(msg + REPOFFSET) = 0;
      *(msg + SCODEOFFSET) = FAILURE;
      strcpy(buf, "No such job or task!");
      bcopy(buf, msg + SMSGOFFSET, strlen(buf) + 1 );
      rpkt->length = SMSGOFFSET + strlen(buf) + 1;
    }
    else {
      ntmp1 = htonl(curjob->jid);
      bcopy(&ntmp1, msg+PRM_JIDOFFSET, LONG_SZ);
      *(msg + REPOFFSET) = HOSTID;
      *(msg + SCODEOFFSET) = SUCCESS;
      bcopy(&h_ad, msg + SMSGOFFSET, PRM_AD_SZ);
      rpkt->length = SMSGOFFSET + PRM_AD_SZ;
    }
    ardp_respond(current_req, ARDP_R_COMPLETE); 
    break;
    
    
  case IJADD_REQ: /* Request from a task for host-addr of its 
		     iotask/jobmngr */
    
    bcopy(msg+NTOFFSET, &req_tid, LONG_SZ);
    req_tid = ntohl(req_tid);
    req_pnum = *(msg+PNUMOFFSET);

    jobfound = FALSE;
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);
    if(curjob) {
      jobfound = TRUE;
      if (req_pnum == 0) /* jobmngr */
	bcopy(&(curjob->jmaddr), &h_ad, PRM_AD_SZ);
      else
	if (req_pnum == 1) /* iotask */
	  bcopy(&(curjob->iotaddr), &h_ad, PRM_AD_SZ);
	else
	  bcopy(&(curjob->fioaddr), &h_ad, PRM_AD_SZ);
    }
    
    /* Port address not known. Make sure that *I*  am not the host
       running the jobmngr; then send a query to the remote node
       on which the jobmngr is executing */
    
    if ( (h_ad.sin_port == 0) && (h_ad.sin_addr.s_addr != inaddr)) { 
      pkt = ardp_ptalloc();   
      msg = pkt->start;       
      
      *(msg+PRM_OP_OFFSET) = (u_char)IJADD_REQ;
      ntmp1 = htonl(req_jid);
      bcopy(&ntmp1, msg + PRM_JIDOFFSET, LONG_SZ);
      bzero(msg+NTOFFSET, LONG_SZ);
      *(msg+PNUMOFFSET) = req_pnum;
      pkt->length = NTOFFSET + LONG_SZ;
      
      h_ad.sin_port = htons((u_short) NODEMNGR_PORT);
      
      newreq = ardp_rqalloc();
      newreq->outpkt = pkt; 
      ardp_send(newreq, 0, &h_ad, -1);
      
      rpkt = newreq->rcvd;
      
      if(rpkt == NOPKT) {     /* Nodemngr did not respond */
	fprintf(stderr, "(%s %s) Address request to remote nodemngr %s timed out!\n", progname, my_hostname, inet_ntoa(h_ad.sin_addr) );
	jobfound = FALSE;  /* return "not-found" to task. Task retries 
			      same query later */
	ardp_rqfree(newreq);
	goto done;
      }
      msg = rpkt->start;
      
      if ((*(msg+REPOFFSET)== HOSTID) && (*(msg+SCODEOFFSET) == SUCCESS))
	bcopy(msg + SMSGOFFSET, &h_ad, PRM_AD_SZ);
      else {
	if(prm_debug)
	  fprintf(stderr, "(%s %s) could not perform tid -> h_addr translation!\n", progname, my_hostname);
      }
      /* cache the port addresses locally */
      if (req_pnum)
	curjob->iotaddr.sin_port = h_ad.sin_port;
      else
	curjob->jmaddr.sin_port = h_ad.sin_port;
      
      ardp_rqfree(newreq);
    }
    
  done:
    rpkt = current_req->outpkt = ardp_ptalloc();
    rpkt->seq = 0;
    msg = rpkt->start;
    
    if ( !jobfound ){
      *(msg + REPOFFSET) = 0;
      *(msg + SCODEOFFSET) = FAILURE;
      strcpy(buf, "No such job or iotask!");
      bcopy(buf, msg+SMSGOFFSET, strlen(buf)+1);
      rpkt->length = SMSGOFFSET+strlen(buf)+1;
    }
    else {
      *(msg + REPOFFSET) = HOSTID;
      *(msg + SCODEOFFSET) = SUCCESS;
      bcopy(&h_ad, msg + SMSGOFFSET, PRM_AD_SZ);
      rpkt->length = SMSGOFFSET + PRM_AD_SZ;
    }
    ardp_respond(current_req, ARDP_R_COMPLETE); 
    break;
    
    
  case NSTATUS_QRY:   /* sysmngr is requesting the busy status of node */
    
    /* save sysmgr's hostaddr */
    bcopy(&(current_req->peer), &smaddr, PRM_AD_SZ); 
    smaddr.sin_family = AF_INET;
    smaddr.sin_port = htons ((u_short)SYSMNGR_PORT);
    rpkt = current_req->outpkt = ardp_ptalloc();
    rpkt->seq = 0;
    msg = rpkt->start;
    *(msg + REPOFFSET) = NSTAT_UPDATE;
    *(msg + SMSGOFFSET) = node_status;
    rpkt->length = SMSGOFFSET + 1;
    ardp_respond(current_req, ARDP_R_COMPLETE);
    break;
    
    
  case PBINDREQ:  /* Request from task to assign a port addr. Increment a
		     local counter, assign that value, and also store it
		     in the local taskinfo structure of the task. */
    
    bcopy(msg + NTOFFSET, &req_tid, LONG_SZ);
    req_tid = ntohl(req_tid);
    req_pnum = *(msg+PNUMOFFSET);
    jobfound = FALSE;
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);
    if(curjob) {
      jobfound = TRUE;
      if (req_tid != 0) {          /* not jobmngr or iotasks */
	SEARCH_LIST(curjob->taskptr,curtask,tid,req_tid);
	if(curtask) {					    
	  if ( req_pnum > curtask->nports) {
	    curtask->nports = req_pnum;
	    curtask->ports = (u_short *)realloc(curtask->ports, 
						req_pnum*2);
	  }
	  *(curtask->ports+req_pnum) = next_udp_port;
	}
      }
      else {  /* jobmngr or tiotask or fiotask */
	if(req_pnum == 0) 
	  curjob->jmaddr.sin_port = htons(next_udp_port);
	else if (req_pnum == 1)
	  curjob->iotaddr.sin_port = htons(next_udp_port);
	else 
	  curjob->fioaddr.sin_port = htons(next_udp_port);
      }
    }
    rpkt = current_req->outpkt = ardp_ptalloc();
    rpkt->seq = 0;
    msg = rpkt->start;
    *(msg + REPOFFSET) = HOSTID;
    if (jobfound) {
      *(msg + SCODEOFFSET) = SUCCESS;
      np = htons(next_udp_port);
      bcopy(&np, msg + SMSGOFFSET, 2);
      rpkt->length = SMSGOFFSET + 2;
      ++next_udp_port;
    }
    else {
      fprintf(stderr, "(%s) No such authorized job %d\n", progname,
	      req_jid);
      *(msg + SCODEOFFSET) = FAILURE;
      rpkt->length = SCODEOFFSET + 1;
    } 
    ardp_respond(current_req, ARDP_R_COMPLETE);
    if (next_udp_port == 0)
      next_udp_port = FIRST_USR_PORT;
    break;
    
    
  case AUTHCODE: /* Sysmngr has sent an authorizaton code for job */
    
    cur_al = (struct authlist *)malloc(sizeof(struct authlist) );
    
    cur_al->jid = req_jid;
    
    bcopy(msg + NTOFFSET, &(cur_al->acode), LONG_SZ);
    /* Authcode is retained in network byte order */
    
    bcopy(msg + NTOFFSET + 4, &(cur_al->addr), PRM_AD_SZ);
    cur_al->addr.sin_family = AF_INET;
    APPEND_ITEM(cur_al, authl_head);
    
    prm_server_respond(current_req, (u_char)SUCCESS, (u_char)SUCCESS,
		       0, (char *)0);
    
    break;
    
    
  case TASKCOUNT: /* Request for number of tasks in this job domain */
    
	  
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);
    if (curjob) 
	ntmp1 = htonl(curjob->numtasks);
    else
      ntmp1 = 0;

    rpkt = current_req->outpkt = ardp_ptalloc();
    rpkt->seq = 0;
    msg = rpkt->start;
    *(msg + REPOFFSET) = TASKCOUNT;
    *(msg + SCODEOFFSET) = SUCCESS;
    bcopy(&ntmp1, msg+SMSGOFFSET, LONG_SZ);
    rpkt->length = SMSGOFFSET + LONG_SZ;
    ardp_respond(current_req, ARDP_R_COMPLETE);
    
    break;
    
    
  case JOBDONE: /* jobmngr notifies nodemngr that job has completed/
		   terminated. nodemngr kills any residual tasks and 
		   deallocates storage for this entry in the job table */
    
    signum = *(msg+SIGOFFSET);
    rpkt = current_req->outpkt = ardp_ptalloc();
    rpkt->seq = 0;
    msg = rpkt->start;
    *(msg + REPOFFSET) = HOSTID;
    *(msg + SCODEOFFSET) = SUCCESS;
    rpkt->length = SCODEOFFSET + 1;
    ardp_respond(current_req, ARDP_R_COMPLETE); /* send ACK to jobmngr */
    
    if (signum == 0) 
      signum = SIGTERM;  /* If termination is not due to a signal from 
			    the user, then use SIGTERM to terminate any
			    stray tasks */
    
    SEARCH_LIST(job_tbl,curjob,jid,req_jid);
    if (!curjob)
      break;
    EXTRACT_ITEM(curjob, job_tbl);
    
    while(curjob->taskptr != NULL) {
      curtask = curjob->taskptr;
      curjob->taskptr = curtask->next;
      cur_pid_ent = curtask->pid_tbl_p;
      /* Check if the task is running on the local node. If so, send
	 signal to that process */
      if (cur_pid_ent)
	if ( (cur_pid_ent->pid > 0) && (curtask->status == T_RUNNING) ) {
	  curtask->status = T_KILLED;
	  kill(cur_pid_ent->pid, signum);
	}
      free(curtask->ports); /* deallocate storage for all port ids desc.*/
      free(curtask);        /* deallocate storage for current task desc.*/
    }
    
    free(curjob);  /* deallocate storage for job entry */
    
    SEARCH_LIST(authl_head,cur_al,jid,req_jid);
    if (cur_al) {
      EXTRACT_ITEM(cur_al, authl_head);
      free(cur_al);
    }
    
    break;
    

 /* Change of busy-status of node, requested by external agent */
  case NSTAT_UPDATE:  
    

    tmp_status = *(msg + PRM_OP_OFFSET + 1);
    prm_server_respond(current_req, (u_char)SUCCESS, (u_char)SUCCESS,
		       0, (char *)0);
    
    if (tmp_status != node_status) {
      NM_inform_sm_node_status(tmp_status);
      node_status = tmp_status;
    }
    break;
    
    
  default:
    break;
  }
}

