/*
 * Copyright (c) 1992, 1993 by the University of Southern California
 *
 * For copying and distribution information, please see the files
 * <prm-copyr.h>.
 *
 * Written by srao 7/92
 *
 */

#include <prm-copyr.h>

#include <ardp.h>

#include <sys/socket.h>
#include <stdio.h>
#include <sys/types.h>

#define  MAIN_PROG
#include <comm.h>

#ifdef HPUX
#   include <signal.h>
#endif

char *progname;


main(int argc, char *argv[], char *envp[])
{
  int     c;
  int     ntasks;
  char    buf[MAX_DIR_LINESIZE];
  char    *progname, *msg;
  u_char  op, my_op;
  u_char  clnt_portnum;
  u_short my_udp_port_num;
  u_long  len, msg_tag;
  u_long  clnt_taskid;
  RREQ    current_req;
  PTEXT   pkt, rpkt;

  extern  void (*sigint_handler)();

  signal(SIGINT, sigint_handler);

  _progname = progname = argv[0];
  sscanf(argv[1], "%lu", &_my_jobid);
  sscanf(argv[2], "%hu", &my_udp_port_num);

  sprintf(buf,"#%d", my_udp_port_num);
  ardp_bind_port(buf);

  /* Jobmngr enforced synchronization of all tasks */

  do {
    current_req = ardp_get_nxt();
  } while( *(current_req->rcvd->start + PRM_OP_OFFSET) != TASK_SYNC);

  current_req->outpkt = ardp_ptalloc();
  *(current_req->outpkt->start + PRM_OP_OFFSET) = (u_char)SUCCESS;
  current_req->outpkt->length = PRM_OP_OFFSET + 1;
  ardp_respond(current_req, ARDP_R_COMPLETE);

  ntasks = numtasks();
  trsln_cache = (struct cache *)calloc((ntasks + 1), sizeof(struct cache));
  bzero(trsln_cache, (ntasks + 1) * sizeof(struct cache));

  /* Here is where the initialization procedure is called. */

  if (TIO_init_procedure() == -1) {
    fprintf(stderr, "Error occured in program initialization\n");
    _jm_host_addr = get_dest_addr(0, 0);
    current_req = ardp_rqalloc();
    current_req->outpkt = ardp_ptalloc();
    *(current_req->outpkt->start + PRM_OP_OFFSET) = APPL_ERR;
    ardp_send(current_req, 0, _jm_host_addr, -1);  /* Send to jobmngr */
  }

  for (;;) {

    current_req = ardp_get_nxt();

    pkt = current_req->rcvd;
    msg = pkt->start;
    op  = *(msg + PRM_OP_OFFSET);

    switch(op) {


    case IO_PRINT:
      
      bcopy ( msg + SNDR_TID_OFF, &clnt_taskid, LONG_SZ);
      clnt_taskid = ntohl(clnt_taskid);
      
      extract_from_pkts(pkt, buf, MAX_DIR_LINESIZE); 
    
      if (clnt_taskid)
	printf("(task %2d) %s\n", clnt_taskid, buf);
      else 
	printf("%s\n", buf);

      if ( ardp_replyf(current_req, ARDP_R_COMPLETE, "#####%c%c", 
		       (u_char)SUCCESS, (u_char)SUCCESS, (char *)0 ) 
	  != ARDP_SUCCESS)
	fprintf(stderr, "(%s) transmit failed.\n", progname);
      
      break;


    case APP_MSG:
      
      bcopy ( msg + SNDR_TID_OFF, &clnt_taskid, LONG_SZ);
      clnt_taskid = ntohl(clnt_taskid);
      
      bcopy (msg + TAG_OFFSET, &msg_tag, LONG_SZ);
      msg_tag = ntohl(msg_tag);

      bcopy(msg + MSGLEN_OFFSET, &len, LONG_SZ);
      len = ntohl(len);
      bcopy(msg + PRM_APPL_MSG_OFFSET, buf, len);

      current_req->outpkt = ardp_ptalloc();
      current_req->outpkt->seq = 0;
      msg = current_req->outpkt->start;
      *(msg + PRM_OP_OFFSET) = SUCCESS;
      *(msg + REPOFFSET) = SUCCESS;
      *(msg + SCODEOFFSET) = SUCCESS;
      current_req->outpkt->length = SCODEOFFSET+1;
      ardp_respond(current_req, ARDP_R_COMPLETE); 
      perform_global_op(clnt_taskid, msg_tag, buf, len);

      break;


    case TASK_SYNC:

      current_req->outpkt = ardp_ptalloc();
      current_req->outpkt->seq = 0;
      msg = current_req->outpkt->start;
      *(msg + REPOFFSET) = SUCCESS;
      *(msg + SCODEOFFSET) = SUCCESS;
      current_req->outpkt->length = SCODEOFFSET+1;
      ardp_respond(current_req, ARDP_R_COMPLETE); 
      break;


    case JOBDONE:

      if ( ardp_replyf(current_req, ARDP_R_COMPLETE, "#####%c%c", 
		       (u_char)SUCCESS, (u_char)SUCCESS) != ARDP_SUCCESS)
	fprintf(stderr, "(%s) transmit failed.\n", progname);
      
      exit(0);
      break;
      
    default:
      fprintf(stderr, "(%s) cannot perform requested operation (code=%d)!\n", 
	      progname, op);
      
      break;
    }
  }
}


void
sigint_handler(sig, code, scp, addr)
int sig, code;
struct sigcontext *scp;
char *addr;
{
  signal(SIGINT, SIG_IGN);
  kill(getppid(), SIGINT);
  signal(SIGINT, sigint_handler);
}

